/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Sat Dec/07/2002 13:48 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef KANAL_H
#define KANAL_H

#ifndef VERBINDUNG_H
#include "verbindung.h"
#endif

namespace fw_cost {

    // --------------------------------------------------------------------------------
    //     class Connection
    // --------------------------------------------------------------------------------
    class Connection {
    private:
        VerbindungPtr verbindung;
        struct tm start; // start of connection
        int duration; // in seconds
        int kanal; // 0 oder 1

    public:
        Connection(VerbindungPtr verbindung_, const struct tm& startTime, int duration_in_seconds, int kanal_) :
            verbindung(verbindung_), start(startTime), duration(duration_in_seconds), kanal(kanal_)
        {
            hassert(kanal==0 || kanal==1);
            assert(duration_in_seconds>0);
        }
        virtual ~Connection() {}

        int getDuration() const { return duration; }
        void setDuration(int newDuration) { duration = newDuration; }
        int getKanal() const { return kanal; }

        const struct tm *getStart() const { return &start; }
        void getEnd(struct tm *t) const {
            *t         = start;
            time_t t1  = mktime(t);
            assert(t1!=-1);

            t1 += getDuration();
            *t  = *localtime(&t1);
        }

        double getCost() const {
            double startCost = verbindung->getCostFor(getStart(), getDuration());
            return startCost;
        }

        const Verbindung& getVerbindung() const { return *verbindung; }

        int operator-(const Connection& other) const { // liefert Differenz der Startzeiten
            struct tm t  = start;
            time_t    t1 = mktime(&t);
            t            = *other.getStart();
            time_t    t0 = mktime(&t);

            return int(difftime(t1, t0));
        }

        int pauseZu(const Connection& other) const {
            int startdiff = other-*this;
            return startdiff-getDuration();
        }

        bool isInZeitraum(const time_t& von, const time_t& bis) const;

        bool operator<(const Connection& other) const {
            int tdiff = *this - other;

            return tdiff<0 || (tdiff==0 && getKanal()<other.getKanal());
        }
    };

    typedef std::list<Connection> ConnectionList;
    typedef ConnectionList::const_iterator ConnIter;

    ConnIter findConnectionSeries(ConnIter begin, ConnIter end, int max_pause, bool allowConnectionChange, bool& usesBothChannels);

    // --------------------------------------------------------------------------------
    //     class Kanal
    // --------------------------------------------------------------------------------
    class Kanal {
    private:
        std::string verbunden_mit;
        struct tm tm;

        static bool correct_date; // einen Tag zu Datum hinzuaddieren (Datumsumschaltung ist fehlgeschlagen)

        int duration(const std::string& datum, const std::string& uhrzeit) const {
            struct tm tm2;
            parseDateTime(&tm2, datum, uhrzeit);
            time_t    t1 = mktime(const_cast<struct tm *>(&tm));
            time_t    t2 = mktime(&tm2);
            return int(difftime(t2, t1));
        }

    public:

        static void parseDateTime(struct tm *t, const std::string& datum, const std::string& uhrzeit);

        bool is_open() const { return verbunden_mit.length(); }
        bool is_closed() const { return !is_open(); }

        static bool date_is_corrected() { return correct_date; }
        static void set_date_corrected(bool b) { correct_date = b; }

        void open(const std::string &verbunden_mit_, const std::string& datum, const std::string& uhrzeit) {
            assert(is_closed());
            verbunden_mit = verbunden_mit_;
            parseDateTime(&tm, datum, uhrzeit);
#if defined(DUMP_DBF_PARSING)
            cout << ">> open " << verbunden_mit << " at " << dateString(&tm) << " " << timeString(&tm) << "\n";
#endif // DUMP_DBF_PARSING
        }

        void close(ConnectionList& connections, VerbindungPtr verbindung, const std::string& datum, const std::string& uhrzeit, int kanal) {
            assert(is_open());
            int Duration = duration(datum, uhrzeit);

            if (Duration<=0) Duration = 1; // fake
            connections.push_back(Connection(verbindung, tm, Duration, kanal));
#if defined(DUMP_DBF_PARSING)
            struct tm tm2;
            parseDateTime(&tm2, datum, uhrzeit);
            cout << ">> close " << verbunden_mit << " at " << dateString(&tm2) << " " << timeString(&tm2) << "\n";
#endif // DUMP_DBF_PARSING

            verbunden_mit = "";
        }
        const std::string& VerbundenMit() const { return verbunden_mit; }

        Kanal() : verbunden_mit("") {}
        virtual ~Kanal() { }
    };

}; // end of namespace fw_cost

#else
#error Kanal.h included twice
#endif // KANAL_H
