/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Sat May/25/2002 14:31 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
/////////////////////////////////////////////////////////////////////////////

#include "Verbindung.h"
#include "fw_cost.h"

using namespace std;
using namespace fw_cost;
using namespace rs;
using namespace rs::err;
using namespace rs::ini;
using namespace rs::date;

//  -----------------------------------------------
//      static int getTwoNum(const char *zahl)
//  -----------------------------------------------
static int getTwoNum(const char *zahl) {
    if (!isdigit(zahl[0]) || !isdigit(zahl[1])) {
        throw Error(string("Zwei Ziffern erwartet in '")+zahl+'\'');
    }
    return (zahl[0]-'0')*10+(zahl[1]-'0');
}
//  ----------------------------------------------------
//      static int scanUhrzeit(const char *uhrzeit)
//  ----------------------------------------------------
static int scanUhrzeit(const char *uhrzeit) {
    // uhrzeit = hh[:mm[:ss]]
    int t = getTwoNum(uhrzeit)*60*60;
    if (uhrzeit[2]) {
        if (uhrzeit[2]!=':') throw Error(string("':' erwartet in '")+uhrzeit+'\'');
        t += getTwoNum(uhrzeit+3)*60;
        if (uhrzeit[5]) {
            if (uhrzeit[5]!=':') throw Error(string("Weiterer ':' erwartet in '")+uhrzeit+'\'');
            t += getTwoNum(uhrzeit+6);
        }
    }
    return t;
}

//  ------------------------------------------------------------------
//      void Verbindungsgebuehr::scanTaktung(const char *taktung)
//  ------------------------------------------------------------------
void Verbindungsgebuehr::scanTaktung(const char *taktung) {
    // taktung = sekunden/takt,cent/min
    char *komma = strchr(taktung, ',');
    if (komma) {
        sekundenProTakt      = atoi(taktung);
        double centProMinute = strtod(komma+1, 0);
        double centProTakt   = centProMinute/60.0*sekundenProTakt;
        euroProTakt          = centProTakt/100.0;
        initialTaktung       = taktung;
        assert(initialTaktung.length());
        return;
    }
    throw rs::err::Error(string("Syntax-Fehler in '")+taktung+'\'');
}
//  --------------------------------------------------------------------
//      void Verbindungsgebuehr::scanZeitraum(const char *zeitraum)
//  --------------------------------------------------------------------
void Verbindungsgebuehr::scanZeitraum(const char *zeitraum) {
    // zeitraum        = [tag]+,hh[:mm[:ss]]-hh[:mm[:ss]]
    const char *komma  = strchr(zeitraum, ',');
    const char *reason = "Unbekannt";

    if (komma) {
        const char *Minus = strchr(komma+1, '-');
        if (Minus) {
            for (int d=0; d<=FEIERTAG; ++d) dayOfWeek[d] = false;
            const char *z = zeitraum;
            while (z<komma) {
                char c = *z++;
                if (c<'0' || c>'7') {
                    throw rs::err::Error(string("Ungueltiges Zeichen '")+c+"' in '"+string(zeitraum, komma)+"'");
                }
                dayOfWeek[c-'0'] = true;
            }

            start = scanUhrzeit(string(komma+1, Minus).c_str());
            stop  = scanUhrzeit(Minus+1);

            initialZeitraum = zeitraum;
            assert(initialZeitraum.length());
            return;
        }
        reason = "Fehlendes Minuszeichen";
    }
    reason = "Fehlendes Komma";
    throw rs::err::Error(string("Syntax-Fehler in '")+zeitraum+"' (Grund: "+reason+")");
}

//  ----------------------------------------------------------------------------------------------------------------
//      double Verbindungsgebuehr::getCostFor(DAY d, int starttime, int seconds, int& seconds_calculated) const
//  ----------------------------------------------------------------------------------------------------------------
double Verbindungsgebuehr::getCostFor(DAY d, int starttime, int seconds, int& seconds_calculated) const {
    assert(seconds>0);
    if (!responsibleFor(d, starttime)) {
        seconds_calculated = 0;
        return 0;
    }

#if defined(DUMP_SINGLE_COST_PARTS)
    cout << "starttime=" << starttime << " start=" << start << " stop=" << stop << "\n";
#endif // DUMP_SINGLE_COST_PARTS
    seconds_calculated = stop-starttime+1;
    if (seconds_calculated>seconds) seconds_calculated = seconds;
#if defined(DUMP_SINGLE_COST_PARTS)
    cout << "seconds_calculated=" << seconds_calculated << "\n";
#endif // DUMP_SINGLE_COST_PARTS
    assert(seconds_calculated>0);
    int takte          = (seconds_calculated/sekundenProTakt)+1;
    seconds_calculated = takte*sekundenProTakt;
#if defined(DUMP_SINGLE_COST_PARTS)
    cout << "takte=" << takte << " seconds_calculated=" << seconds_calculated << "\n";
#endif // DUMP_SINGLE_COST_PARTS
    return euroProTakt*takte;
}

//  -----------------------------------------------------------------------------
//      int Verbindung::secondsToNextTarifChange(DAY d, int starttime) const
//  -----------------------------------------------------------------------------
int Verbindung::secondsToNextTarifChange(DAY d, int starttime) const {
    int next_start = INT_MAX;

    for (list<Verbindungsgebuehr>::const_iterator gebuehr=gebuehren.begin(); gebuehr!=gebuehren.end(); ++gebuehr) {
        if (gebuehr->responsibleFor(d)) {
            int start = gebuehr->get_start();
            if (start>=starttime && start<next_start) {
                next_start = start;
            }
        }
    }

    return next_start==INT_MAX ? 0 : next_start-starttime;
}

const Verbindungsgebuehr *Verbindung::lastUsed = 0;

//  ---------------------------------------------------------------------------------
//      double Verbindung::getCostFor(const struct tm *start, int seconds) const
//  ---------------------------------------------------------------------------------
double Verbindung::getCostFor(const struct tm *start, int seconds) const {
    double cost       = 0;
    DAY d             = DAY(Date(*start));
    int sinceMidnight = (start->tm_hour*60+start->tm_min)*60+start->tm_sec;

    while (seconds>0) {
        int tillNextChange = secondsToNextTarifChange(d, sinceMidnight);
#if defined(DEBUG)
        int old_seconds    = seconds;
#endif // DEBUG

        for (list<Verbindungsgebuehr>::const_iterator gebuehr=gebuehren.begin();
             gebuehr != gebuehren.end() && seconds>0;
             ++gebuehr)
        {
            int done;
            int seconds_to_calc = (tillNextChange && tillNextChange<seconds) ? tillNextChange : seconds;

            cost += gebuehr->getCostFor(d, sinceMidnight, seconds_to_calc, done);
#if defined(DUMP_SINGLE_COST_PARTS)
            cout << "cost=" << cost << " done=" << done << "\n";
#endif // DUMP_SINGLE_COST_PARTS
            if (done) {
                lastUsed = &*gebuehr;
                if (done>seconds) done=seconds; // passiert bei taktung > 1 sec
                seconds -= done;
                sinceMidnight += done;
                if (sinceMidnight>=SECONDS_PER_DAY) {
                    sinceMidnight -= SECONDS_PER_DAY;
                    d = d.next();
                }
                break; // restart loop through Verbindungsgebuehr-list
            }
        }
#if defined(DEBUG)
        if (old_seconds == seconds) {
            cerr << "keine passende Verbindungsgebuehr gefunden\n";
            assert(0);
        }
#endif // DEBUG
    }
    return cost;
}

namespace fw_cost {

    // --------------------------------------------------------------------------------
    //     static string ask_User(const string& question)
    // --------------------------------------------------------------------------------
    string ask_User(const string& question) {
        string answer;
        while (1) {
            cerr << question << endl;
            getline(cin, answer, '\n');
            if (answer.length()) break;
        };

        return answer;
    }

    // --------------------------------------------------------------------------------
    //     VerbindungenPtr getVerbindungen(InifilePtr ini)
    // --------------------------------------------------------------------------------
    VerbindungenPtr getVerbindungen(InifilePtr ini) {
        VerbindungenPtr verb(new Verbindungen);
        SectionPtr verbindungen = ini->findSection("Verbindungen");

        if (!verbindungen.Null()) {
            for (Entries::const_iterator e=verbindungen->begin(); e!=verbindungen->end(); ++e) {
                const EntryPtr se = e->second; // dynamic_cast<const StringEntry*>(&*(e->second));
//                 if (!se) throw Error("Numerischer Eintrag in [Verbindungen]");

                const string& verbindungs_name = se->getContent();
                SectionPtr verbindung          = ini->findSection(verbindungs_name);

                if (verbindung.Null()) {
                    throw Error(string("Unbekannte Verbindung '")+verbindungs_name+"' in [Verbindungen]");
                }

                VerbindungPtr v(new Verbindung(verbindungs_name));
                (*verb)[verbindungs_name] = v;

                EntryPtr faktorEntry = verbindung->findEntry("Faktor");
                if (!faktorEntry.Null()) {
                    double faktor = faktorEntry->getNumericContent();
                    v->setFaktor(faktor);
                }

                for (unsigned num=0; ;++num) { // Suche Taktungen
                    string entry_name          = getTaktungEntryName(num);
                    string zeitraum_entry_name = entry_name+"_Zeitraum";
                    string taktung_entry_name  = entry_name+"_Taktung";

                    EntryPtr zeitraum = verbindung->findEntry(zeitraum_entry_name);
                    if (zeitraum.Null()) break;

                    EntryPtr taktung = verbindung->findEntry(taktung_entry_name);
                    if (taktung.Null()) {
                        throw Error(string("Fehlender Eintrag ")+taktung_entry_name+" in Section "+verbindungs_name);
                    }

//                     const StringEntry *zeitraum_se = dynamic_cast<const StringEntry*>(&*zeitraum);
//                     const StringEntry *taktung_se  = dynamic_cast<const StringEntry*>(&*taktung);

//                     if (!zeitraum_se) throw Error(string("Format-Fehler in ")+zeitraum_entry_name);
//                     if (!taktung_se) throw Error(string("Format-Fehler in ")+taktung_entry_name);

                    try {
                        v->addGebuehr(taktung->getContent(), zeitraum->getContent());
                    }
                    catch (Error& err) {
                        throw Error(string("Fehler in Taktung: ")+err.get_message());
                    }
                }
            }
        }

        return verb;
    }

    // --------------------------------------------------------------------------------
    //     VerbindungPtr findVerbindung(VerbindungenPtr verb, const string& verbunden_mit, bool& ini_changed)
    // --------------------------------------------------------------------------------
    VerbindungPtr findVerbindung(VerbindungenPtr verb, const string& verbunden_mit, bool& ini_changed)  {
        Verbindungen::iterator verbindung = verb->find(verbunden_mit);
        if (verbindung==verb->end()) { // unbekannte Verbindung
            cerr << "\nUnbekannte Verbindung zu '" << verbunden_mit << '\'' << "\n";

            VerbindungPtr v(new Verbindung(verbunden_mit.c_str()));
            (*verb)[verbunden_mit] = v;

            cerr << "\nFormat fuer Taktung : Sekunden/Takt,Cent/min"
                 << "\n                      Bsp.: 60,1.15" << endl;

            int err_count = 0;

#if defined(AUTO_ANSWER)
            v->addGebuehr("1, 0.5", "0123456,00-23:59:59");
#else
            {
                bool error = false;
                do {
                    error         = false;
                    string answer = ask_User("Meist benoetigte Taktung?");

                    try {
                        v->addGebuehr(answer, "01234567,00-23:59:59");
                    }
                    catch (Error& err) {
                        if (++err_count==5) throw Error("Zu viele falsche Benutzer-Eingaben.");
                        cerr << "\nFehler in Eingabe: " << err.get_message() << "\n"
                             << "    Beispiel-Antwort:   60,1.25\n"
                             << "    Beispiel-Bedeutung: Minutentaktung mit 1.25 Cent/min" << endl;
                        error = true;
                    }
                } while (error);
            }
#endif // AUTO_ANSWER

            cerr << "\nSpaeter definierte Zeitraeume und Taktungen ueberschreiben eher definierte.\n"
                 << "Die Eingabe von 'e' beendet die Definition von weiteren Zeitraeumen.\n"
                 << "\n"
                 << "Format fuer Zeitraum: [tag]+,hh[:mm[:ss]]-hh[:mm[:ss]]\n"
                 << "                      (tag: 0=Sonntag, 1=Montag, ..., 6=Samstag, 7=Feiertag\n"
                 << "                            a=alle w=werktage f=Samstag+Sonntag+Feiertag)" << endl;

            bool enough = false;
            err_count   = 0;
            while (!enough) {
                bool error = false;
                string zeitraum;
                string taktung;

                do {
                    error = false;
#if defined(AUTO_ANSWER)
                    zeitraum = "e";
#else
                    zeitraum = ask_User("\nWeiterer Zeitraum?");
#endif // AUTO_ANSWER
                    if (zeitraum=="e") { enough = true; break; }

                    {
                        string shortcut = zeitraum.substr(0, 2);
                        if (shortcut == "a,") zeitraum = "01234567"+zeitraum.substr(1);
                        else if (shortcut == "w,") zeitraum = "12345"+zeitraum.substr(1);
                        else if (shortcut == "f,") zeitraum = "067"+zeitraum.substr(1);
                    }

                    try { Verbindungsgebuehr test("60,1", zeitraum); }
                    catch (Error& err) {
                        if (++err_count==5) throw Error("Zu viele falsche Benutzer-Eingaben.");
                        cerr << "\nFehler in Eingabe: " << err.get_message() << "\n"
                             << "    Beispiel-Antwort:   12345, 09:00-18:00\n"
                             << "    Beispiel-Bedeutung: werktags, 9:00 bis 18:00 Uhr" << endl;
                        error = true;
                    }
                } while (error);

                if (enough) break;

                err_count = 0;
                do {
                    error   = false;
                    taktung = ask_User(string("\nTaktung fuer Zeitraum '")+zeitraum+"'?");
                    if (taktung=="ende") { enough = true; break; }

                    try { v->addGebuehr(taktung, zeitraum); }
                    catch (Error& err) {
                        if (++err_count==5) throw Error("Zu viele falsche Benutzer-Eingaben.");
                        cerr << "\nFehler in Eingabe: " << err.get_message() << "\n"
                             << "    Beispiel-Antwort:   60,1.5\n"
                             << "    Beispiel-Bedeutung: Minutentaktung mit 1.5 Cent/min" << endl;
                        error = true;
                    }
                } while (error);
            }

            verbindung = verb->find(verbunden_mit);
            assert(verbindung!=verb->end());

            ini_changed = true;
        }

        if (verbindung==verb->end())
            throw InternalError(string("Unbekannte Verbindung zu '")+verbunden_mit+'\'');

        return verbindung->second;
    }

}; // end of namespace fw_cost



