/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Sat Dec/07/2002 13:47 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
/////////////////////////////////////////////////////////////////////////////

#ifndef VERBINDUNG_H
#define VERBINDUNG_H

#ifndef __LIST__
#include <list>
#endif

#ifndef INIFILE_H
#include <inifile.h>
#endif
#ifndef DATE_H
#include "Date.h"
#endif


namespace fw_cost {

    class DAY {
    private:
        rs::date::Date d;

    public:
        DAY(rs::date::Date d_) : d(d_) {}
        virtual ~DAY() {}

        DAY next() const {
            rs::date::Date d2(d);
            d2.inc();
            return DAY(d2);
        }

        int dayOfWeek() const { return d.dayOfWeek(); }
        bool is_Feiertag() const {
            int idx     = d.is_Feiertag();
            return idx != 0 && rs::date::is_catholic(idx) == false;
        }
    };


//     // ----------------------
//     //     typedef enum
//     // ----------------------
//     typedef enum {
//         SONNTAG=0, MONTAG, DIENSTAG, MITTWOCH, DONNERSTAG, FREITAG, SAMSTAG,
//         DAYS, FEIERTAG
//     } DAY;

//     //  ----------------------------------
//     //      inline DAY nextDAY(DAY d)
//     //  ----------------------------------
//     inline DAY nextDAY(DAY d) {
//         return DAY((int(d)+1)%DAYS);
//     }

//     //  ---------------------------------------------
//     //      inline const DAY& operator++(DAY& d)
//     //  ---------------------------------------------
//     inline const DAY& operator++(DAY& d) {
//         d = DAY(int(d)+1);
//         return d;
//     }

#define SECONDS_PER_DAY (24*60*60)

#define FEIERTAG 7 // index of Feiertag in Verbindungsgebuehr::dayOfWeek

    // --------------------------------------------------------------------------------
    //     class Verbindungsgebuehr
    // --------------------------------------------------------------------------------
    class Verbindungsgebuehr {
    private:
        bool dayOfWeek[8];
        int start, stop; // in Sekunden seit o:oo Uhr

        double euroProTakt;
        int sekundenProTakt;

        std::string initialZeitraum;
        std::string initialTaktung;

        void scanTaktung(const char *taktung);
        void scanZeitraum(const char *zeitraum);

    public:
        Verbindungsgebuehr(const std::string& taktung, const std::string& zeitraum)
        {
            scanTaktung(taktung.c_str());
            scanZeitraum(zeitraum.c_str());
        }
        virtual ~Verbindungsgebuehr() {}

        bool responsibleFor(DAY d) const { return dayOfWeek[d.dayOfWeek()] || (dayOfWeek[FEIERTAG] && d.is_Feiertag()); }
        int get_start() const { return start; }

        bool responsibleFor(DAY d, int starttime) const {
            return starttime>=start && starttime<=stop && responsibleFor(d);
        }

        int getTakt() const { return sekundenProTakt; }

        double getCostFor(DAY d, int starttime, int seconds, int& seconds_calculated) const;
        const std::string& getZeitraum() const { assert(initialZeitraum.length()); return initialZeitraum; }
        const std::string& getTaktung() const { assert(initialTaktung.length()); return initialTaktung; }
    };

    // --------------------------------------------------------------------------------
    //     class Verbindung
    // --------------------------------------------------------------------------------
    class Verbindung {
    private:
        std::string name;
        std::string shortName;
        double faktor; // faktor fuer Kostenabschtzung

        std::list<Verbindungsgebuehr> gebuehren; // spezielle Tarifzeiten am Anfang der Liste; allgemeine am Schluss

        static const Verbindungsgebuehr *lastUsed; // Verbindungsgebuehr die beim letzten Aufruf von getCostFor als letztes benutzt wurde

    public:
        Verbindung(const std::string& name_) :
            name(name_),
            faktor(1.0)
        {
            size_t kl_auf = name.find_first_of("([");
            if (kl_auf==std::string::npos) {
                shortName = name;
            }
            else {
                shortName = rs::str::cropString(name.substr(0, kl_auf));
            }
        }
        virtual ~Verbindung() {}

        void setFaktor(double faktor_) { faktor = faktor_; }
        double getFaktor() const { return faktor; }

        static const Verbindungsgebuehr *getLastUsed() { return lastUsed; }

        void addGebuehr(const std::string& taktung, const std::string& zeitraum) {
            gebuehren.push_front(Verbindungsgebuehr(taktung, zeitraum));
        }

        std::list<Verbindungsgebuehr>::const_reverse_iterator rbegin() const { return gebuehren.rbegin(); }
        std::list<Verbindungsgebuehr>::const_reverse_iterator rend() const { return gebuehren.rend(); }

        const std::string& getName() const { return name; }
        const std::string& getShortName() const { return shortName; }

        int secondsToNextTarifChange(DAY d, int starttime) const;
        double getCostFor(const struct tm *start, int seconds) const;
    };

    // --------------------------------------------------------------------------------
    //     class Verbindungen
    // --------------------------------------------------------------------------------

    typedef rs::SmartPtr<Verbindung> VerbindungPtr;
    typedef std::map<std::string, VerbindungPtr, std::less<std::string> > Verbindungen; // string enthaelt langen Verbindungsnamen
    typedef rs::SmartPtr<Verbindungen> VerbindungenPtr;

    VerbindungenPtr getVerbindungen(rs::ini::InifilePtr ini);
    VerbindungPtr findVerbindung(VerbindungenPtr verb, const std::string& verbunden_mit, bool& ini_changed);

    std::string ask_User(const std::string& question);

}; // end of namespace fw_cost

#else
#error Verbindung.h included twice
#endif // VERBINDUNG_H
