/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Sat Dec/07/2002 17:19 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
/////////////////////////////////////////////////////////////////////////////

#include <set>
#include <sstream>
#include <vector>

#include "fw_cost.h"
#include "output.h"

using namespace std;
using namespace rs;
using namespace rs::str;
using namespace fw_cost;
using namespace rs_boost;

namespace fw_cost {
    //  ----------------------------------------------
    //      string timeString(const struct tm *t)
    //  ----------------------------------------------
    string timeString(const struct tm *t) {
#define BUFSIZE 100
        char buffer[BUFSIZE+1];
        strftime(buffer, BUFSIZE, "%H:%M:%S", t);
        return string(buffer);
#undef BUFSIZE
    }
    //  ---------------------------------------
    //      string timeString(int seconds)
    //  ---------------------------------------
    string timeString(int seconds) {
        if (seconds<60) return str::strf("        %02li s", seconds);
        if (seconds<(60*60)) return str::strf("     %02li:%02li m", seconds/60, seconds%60);
        return str::strf("%4s:%02li:%02li h", str::strf("%02i", seconds/(60*60)).c_str(), (seconds%(60*60))/60, seconds%60);
    }
};
// --------------------------------------------------------------------------------
//      static string doubleString(double d, int stellen, int vorkommastellen)
// --------------------------------------------------------------------------------
static string doubleString(double d, int stellen, int vorkommastellen) {
    int    faktor       = 1;
    int    s            = stellen;
    while (s--) faktor *= 10;
    string a            = str::strf("%i", int(d*faktor+0.5));

    while (a.length()<2) a = string(" ")+a;

    a = a.substr(0, a.length()-2)+'.'+a.substr(a.length()-2);
    a = "       "+a;
    a = a.substr(a.length()-(vorkommastellen+stellen+1));

    return a;
}
//  ---------------------------------------------
//      static string preisString(double euro)
//  ---------------------------------------------
static string preisString(double euro) {
    if (euro>=1.0) return doubleString(euro, 2, 4)+" Euro";
    return doubleString(euro*100, 2, 4)+" Cent";
}


namespace fw_cost {

    //  ----------------------
    //      class Summary
    //  ----------------------
    class Summary {
    private:
        int    realTime;
        int    virtualTime;     // incl. Pausen
        double kosten;
        double weightedCostPerMinute; // cost per minute * duration
        int    connections;
        int    connections2;    // Mehrfach-Verbindungen einfach

    public:
        Summary()
            : realTime(0), virtualTime(0), kosten(0), weightedCostPerMinute(0), connections(0), connections2(0)
        {}
        virtual ~Summary() {}

        void print(ostream& out, const string& title) const {
            out << title << ':' << '\n';
            if (kosten) {
                out << '\n'
                    << "Gesamt-Onlinezeit (real)    : " << timeString(realTime) << " (=" << realTime/60 << "min; =" << realTime << "sec)" << '\n'
                    << "Gesamt-Onlinezeit (virtuell): " << timeString(virtualTime) << " (=" << virtualTime/60 << "min; =" << virtualTime << "sec)" << '\n'
                    << "Gesamt-Kosten               : " << preisString(kosten) << '\n'
                    << "Kosten pro Minute (real)    : " << preisString((kosten/realTime)*60) << '\n'
                    << "Kosten pro Minute (virtuell): " << preisString(weightedCostPerMinute/virtualTime) << '\n'
                    << "Verbindungen (real)         : " << connections << '\n'
                    << "Verbindungen (virtuell)     : " << connections2 << '\n'
                    << '\n';
            }
            else {
                out << "Keine Verbindungen.\n\n";
            }
        }

        void addSingle(int duration, double cost) {
            realTime              += duration;
            virtualTime           += duration;
            kosten                += cost;
            weightedCostPerMinute += cost*60; // pro_minute = (cost/duration)*60; weighted = pro_minute*duration
            connections++;
            connections2++;
        }

        void addMulti(int duration_real, int duration_virtual, double cost, int single_connections) {
            realTime              += duration_real;
            virtualTime           += duration_virtual;
            kosten                += cost;
            weightedCostPerMinute += cost*60;
            connections           += single_connections;
            connections2++;
        }
    };

    typedef map<string, Summary, less<string> > SummaryMap;

    //  ------------------
    //      class Sum
    //  ------------------
    class Sum {
    private:
        string text;
        double value;           // cost or time

    public:
        Sum(const string& text_, double value_) : text(text_), value(value_) {}
        virtual ~Sum() {}

        double Value() const { return value; }
        const string& Text() const { return text; }
    };


    // --------------------------------------------------------------------------------------------------
    //     void printConnectionSums(const VerbindungenPtr verb, const ConnectionList& connections,
    // --------------------------------------------------------------------------------------------------
    void printConnectionSums(const VerbindungenPtr verb, const ConnectionList& connections, const time_t& von, const time_t& bis, const string& zeitraum) {
        SummaryMap summaries;

        for (ConnIter c=connections.begin(); c != connections.end(); ++c) {
            if (c->isInZeitraum(von, bis)) {
                summaries[c->getVerbindung().getName()].addSingle(c->getDuration(), c->getCost());
            }
        }

        for (Verbindungen::const_iterator v=verb->begin(); v != verb->end(); ++v)
        {
            const string& name = v->first;
            summaries[name].print(cout, string("Verbindungen zu '")+name+"' (Zeitraum: "+zeitraum+")");
        }
        cout << '\n';
    }

    //  -----------------------------------------------------------------------------------------------------------------------------------------------------------
    //      void interpretConnections(const ConnectionList& connections, int max_pause, bool allowConnectionChange, bool einzel, bool detailed, bool provider)
    //  -----------------------------------------------------------------------------------------------------------------------------------------------------------
    void interpretConnections(const ConnectionList& connections, int max_pause, bool allowConnectionChange, bool einzel, bool detailed, bool provider) {
        ConnIter   first = connections.begin();
        Summary    summary;
        SummaryMap single_summaries;
        string     SEP   = "  |  ";

        if (einzel) section_______________________("Einzelverbindungen");

        while (first!=connections.end()) {
            ConnIter last;
            bool     usesBothChannels = false;
            if (max_pause) {
                last = findConnectionSeries(first, connections.end(), max_pause, allowConnectionChange, usesBothChannels);
            }
            else {
                last = first;
            }

            if (last==first) {  // einzelne Verbindung
                double cost       = first->getCost();
                int    duration   = first->getDuration();
                double pro_minute = (cost/duration)*60;

                if (einzel) {
                    cout << "- " << dateString(first->getStart()) << " " << timeString(first->getStart()) << SEP
                         << (first->getKanal() ? "2 " : "  ")
                         << "Dauer: " << timeString(duration) << "             " << SEP
                         << "Kosten: " << preisString(cost) << SEP
                         << "p/min: " << preisString(pro_minute) << SEP
                         << first->getVerbindung().getShortName()
                         << '\n';
                }

                summary.addSingle(duration, cost);
                single_summaries[first->getVerbindung().getName()].addSingle(duration, cost);
            }
            else {
                ConnIter    after_last   = last; ++after_last;
                double      gesamtKosten = 0;
                int         gesamtDauer  = int(*last-*first)+last->getDuration();
                int         realDauer    = 0; // ohne pausen
                int         verbindungen = 0;
                set<string> beteiligt;
                ConnIter    c            = first;

                while (c!=after_last) {
                    double        kosten    = c->getCost();
                    ConnIter      c_next    = next(c);
                    int           duration  = c->getDuration();
                    const string& shortName = c->getVerbindung().getShortName();

                    beteiligt.insert(shortName);

                    if (detailed) {
                        cout << "  " << dateString(c->getStart()) << " " << timeString(c->getStart()) << SEP
                             << (c->getKanal() ? "2 " : "  ")
                             << "Dauer: " << timeString(duration) << "             " << SEP
                             << "Kosten: " << preisString(kosten) << SEP
                             << shortName;

                        //                     cout << "M " << shortName
                        //                          << " (" << dateString(c->getStart()) << " " << timeString(c->getStart())
                        //                          << ") Dauer: " << timeString(duration) << " Kosten: "
                        //                          << preisString(kosten);
                        if (c!=last) {
                            int pause = c->pauseZu(*c_next);
                            cout << SEP << "Pause: " << timeString(pause);
                        }
                        cout << '\n';
                    }

                    //single_summaries[c->getVerbindung().getName()].addSingle(duration, kosten);

                    gesamtKosten += kosten;
                    assert(duration>=0);
                    realDauer    += duration;
                    verbindungen++;

                    c = c_next;
                }

                double hypoKosten = first->getVerbindung().getCostFor(first->getStart(), gesamtDauer);
                double ersparnis  = hypoKosten-gesamtKosten;

                if (ersparnis<0.00001 && ersparnis>-0.00001) ersparnis = 0;

                double pro_minute = (gesamtKosten/gesamtDauer)*60.0;

                if (einzel) {
                    if (realDauer<0) {
                        assert(0);
                    }

                    cout << "+ " << dateString(first->getStart()) << " " << timeString(first->getStart()) << SEP
                         << (usesBothChannels ? "2 " : "  ")
                         << "Dauer: " << timeString(gesamtDauer) << " " << timeString(realDauer) << SEP
                         << "Kosten: " << preisString(gesamtKosten) << SEP
                         << "p/min: " << preisString(pro_minute) << SEP;

                    // cout << "M " << dateString(first->getStart()) << ' ' << timeString(first->getStart()) << ' ';
                    for (set<string>::iterator b=beteiligt.begin(); b!=beteiligt.end(); ++b) {
                        if (b!=beteiligt.begin()) cout << '/';
                        cout << *b;
                    }
                    cout << '\n';
                    //                 cout << " Dauer: " << timeString(gesamtDauer) << " (=" << preisString(hypoKosten) << ") "
                    //                      << " Online: " << timeString(realDauer)
                    //                      << " Kosten: " << preisString(gesamtKosten)
                    //                     //                 << " Ersparnis: " << preisString(ersparnis) << '\n'
                    //                      << " pro Minute: " << preisString(pro_minute)
                    //                      << '\n';
                }

                summary.addMulti(realDauer, gesamtDauer, gesamtKosten, verbindungen);
                single_summaries[first->getVerbindung().getName()].addMulti(realDauer, gesamtDauer, gesamtKosten, verbindungen);
            }

            ++last;
            first = last;
        }
        cout << '\n';

        section_______________________("Gesamtstatistik:");
        summary.print(cout, string("Alle Verbindungen"));
        if (provider) {
            section_______________________("Gesamtstatistik (nach Providern getrennt):");
            for (SummaryMap::iterator sm=single_summaries.begin(); sm!=single_summaries.end(); ++sm) {
                sm->second.print(cout, string("Verbindungen zu ")+sm->first);
            }
        }

        cout << "\n(virtuell) bezieht sich hierbei auf mehrere hintereinanderfolgende Verbindungen zwischen \n"
            "denen nur kurze Pausen waren. Wenn man die Lnge der Pause auf -1 sec stellt (mit -p-1),  dann \n"
            "sind die Werte fr (real) und (virtuell) identisch.\n";
    }

    //  -------------------------------------------------------------------------------------
    //      inline bool isForProvider(const string *forProvider, const Connection& conn)
    //  -------------------------------------------------------------------------------------
    inline bool isForProvider(const string *forProvider, const Connection& conn) {
        return forProvider==0 || conn.getVerbindung().getShortName()==*forProvider;
    }

    // ----------------------------------
    //      struct data_per_category
    // ----------------------------------
    struct data_per_category {
        double cost;
        double seconds;

        data_per_category() : cost(0), seconds(0) {}
        double price() const { return cost*60.0/seconds; }
    };

    typedef vector<data_per_category> CategoryVector;

#define BREITE 40

    // ---------------------------------------------------------------------------------------------------------------------------------------------------------
    //      static void printCategorizedSums_internal(ostream& out, const ConnectionList& connections, const string& summenName, sum_categorizer categorize,
    // ---------------------------------------------------------------------------------------------------------------------------------------------------------
    static void printCategorizedSums_internal(ostream& out, const ConnectionList& connections, const string& summenName, sum_categorizer categorize,
                                             category_prefix_writer writePrefix, const string *forProvider, set<string> *collectProviders)
    {
        out << '\n' << summenName << ':' << '\n';
        CategoryVector categories;

        for (ConnectionList::const_iterator ci = connections.begin(); ci != connections.end(); ++ci) {
            const string& providerName = ci->getVerbindung().getShortName();
            if (collectProviders) collectProviders->insert(providerName);

            if (forProvider == 0 || *forProvider == providerName) {
                size_t category = categorize(*ci);
                if (category >= categories.size()) categories.resize(category+1);

                categories[category].cost    += ci->getCost();
                categories[category].seconds += ci->getDuration();
            }
        }

        double mean_cost  = 0.0;
        double mean_time  = 0.0;
        double mean_price = 0.0;

        double max_cost  = 0.0;
        double max_time  = 0.0;
        double max_price = 0.0;

        for (CategoryVector::const_iterator cat = categories.begin(); cat != categories.end(); ++cat) {
            mean_cost  += cat->cost;
            mean_time  += cat->seconds;
            mean_price += cat->price();
            max_cost    = gen_algo::max(max_cost, cat->cost);
            max_time    = gen_algo::max(max_time, cat->seconds);
            max_price   = gen_algo::max(max_price, cat->price());
        }

        mean_cost  /= categories.size();
        mean_time  /= categories.size();
        mean_price /= categories.size();

        int mean_cost_width = int((BREITE*mean_cost)/max_cost + 0.5);
        int mean_time_width = int((BREITE*mean_time)/max_time + 0.5);
        int mean_price_width = int((BREITE*mean_price)/max_price + 0.5);

        int category = 0;
        for (CategoryVector::const_iterator cat = categories.begin(); cat != categories.end(); ++cat, ++category) {
            int    d;
            double price = cat->price();

            // print costs :
            writePrefix(out, category);
            out << ' ' << preisString(cat->cost) << ' ';
            int dollars = int((BREITE*cat->cost)/max_cost + 0.5);
            for (d=0; d<dollars && d<mean_cost_width; ++d) out << 'o';
            for (; d<dollars; ++d) out << 'O';
            for (; d<BREITE+1; ++d) out << ' ';

            // print time:
            int zzzzs = int((BREITE*cat->seconds)/max_time + 0.5);
            out << timeString(int(cat->seconds))  << ' ';
            for (d=0; d<zzzzs && d<mean_time_width; ++d) out << 'z';
            for (; d<zzzzs; ++d) out << 'Z';
            for (; d<BREITE+1; ++d) out << ' ';

            // print price:
            int pppps = int((BREITE*price)/max_price + 0.5);
            out << preisString(price) << "/min ";
            for (d=0; d<pppps && d<mean_price_width; ++d) out << 'p';
            for (; d<pppps; ++d) out << 'P';
            out << '\n';
        }
    }

    // -------------------------------------------------------------------------------------------------------------------------------------------
    //      static void printSums_internal(ostream& out, const ConnectionList& connections, const string& summenName, sum_predicate predicate,
    // -------------------------------------------------------------------------------------------------------------------------------------------
    static void printSums_internal(ostream&       out, const ConnectionList& connections, const string& summenName, sum_predicate predicate,
                                   prefix_writer  writePrefix, double& meanValue, int extraLastSum1, int extraLastSum2,
                                   const string  *forProvider, set<string> *collectProviders)
    {

        out << '\n' << summenName << ':' << '\n';

        ConnectionList::const_iterator c1 = connections.begin();
        ConnectionList::const_iterator c2 = next(c1);

        double    overall_cost       = 0;
        double    overall_time       = 0;
        double    overall_price      = 0;
        double    last_cost1         = 0; // von den extraLastSum1 Tagen/Wochen/Monaten
        double    last_time1         = 0;
        double    last_cost2         = 0; // von den extraLastSum2 Tagen/Wochen/Monaten
        double    last_time2         = 0;
        double    cost               = 0;
        double    seconds            = 0;
        double    max_cost           = -1;
        double    max_time           = -1;
        double    max_price          = -1;
        double    mean_cost          = -1;
        double    mean_time          = -1;
        double    mean_price         = 0;
        double    firstAndLast_cost  = 0; // first and last sum (they may be partial)
        double    firstAndLast_time  = 0; // first and last sum (they may be partial)
        double    firstAndLast_price = 0; // first and last sum (they may be partial)
        list<Sum> cost_summen;
        list<Sum> time_summen;
        list<Sum> price_summen;

        if (c1!=connections.end()) {
            if (isForProvider(forProvider, *c1)) {
                cost    += c1->getCost();
                seconds += c1->getDuration();
            }

            for (; ; ++c1,++c2) {
                if (collectProviders) collectProviders->insert(c1->getVerbindung().getShortName());
#if defined(DEBUG)
//                 cout << "  " << dateString(c1->getStart()) << " " << timeString(c1->getStart())
//                      << " " << "Dauer: " << timeString(c1->getDuration()) << '\n';
#endif // DEBUG

                bool last = c2==connections.end();

                if (!last && predicate(*c1, *c2)) {
                    if (isForProvider(forProvider, *c2)) {
                        cost += c2->getCost();
                        seconds += c2->getDuration();
                    }
                }
                else {
                    ostringstream out_time;
                    ostringstream out_cost;
                    ostringstream out_price;

                    double price = (cost*60.0)/seconds;

                    assert(cost>0);
                    assert(seconds>0);
                    assert(price>0);

                    if (firstAndLast_cost==0) firstAndLast_cost = cost;
                    if (firstAndLast_time==0) firstAndLast_time = seconds;
                    if (firstAndLast_price==0) firstAndLast_price = price;

                    if (cost!=0 && seconds!=0) {
                        writePrefix(out_cost, *c1);
                        writePrefix(out_time, *c1);
                        writePrefix(out_price, *c1);

                        out_cost << preisString(cost);
                        out_time << timeString((int)seconds);
                        out_price << preisString(price) << "/min";

                        overall_cost += cost;
                        overall_time += seconds;
                        overall_price += price;

                        max_cost  = gen_algo::max(cost, max_cost);
                        max_time  = gen_algo::max(seconds, max_time);
                        max_price = gen_algo::max(price, max_price);

                        mean_cost  += cost;
                        mean_time  += seconds;
                        mean_price += price;
#if defined(DEBUG)
                        //cout << out_cost.str() << " " << out_time.str() << '\n';
#endif                          // DEBUG

                        cost_summen.push_back(Sum(out_cost.str(), cost));
                        time_summen.push_back(Sum(out_time.str(), seconds));
                        price_summen.push_back(Sum(out_price.str(), price));
                    }

                    if (last) {
                        firstAndLast_cost += cost;
                        firstAndLast_time += seconds;
                        firstAndLast_price += price;
                        break;
                    }

                    if (isForProvider(forProvider, *c2)) {
                        cost    = c2->getCost();
                        seconds = c2->getDuration();
                    }
                    else {
                        cost    = 0;
                        seconds = 0;
                    }
                }
            }
        }

        mean_cost  /= cost_summen.size();
        mean_time  /= time_summen.size();
        mean_price /= price_summen.size();

        assert(mean_price >= 0);
        assert(max_price >= mean_price);

        // Balken an die einzelnen Zeilen anhngen :

        list<Sum>::iterator sc = cost_summen.begin();
        list<Sum>::iterator st = time_summen.begin();
        list<Sum>::iterator sp = price_summen.begin();

        int rest      = cost_summen.size();
        if ((rest-2) <= extraLastSum1) extraLastSum1 = rest-2; // ohne ersten und letzten
        if ((rest-2) <= extraLastSum2) extraLastSum2 = rest-2; // ohne ersten und letzten

        int mean_cost_width = int((BREITE*mean_cost)/max_cost + 0.5);
        int mean_time_width = int((BREITE*mean_time)/max_time + 0.5);
        int mean_price_width = int((BREITE*mean_price)/max_price + 0.5);

        while (sc!=cost_summen.end()) {
            int d;

            // print costs:
            int dollars = int((BREITE*sc->Value())/max_cost + 0.5);
            out << sc->Text() << ' ';
            for (d=0; d<dollars && d<mean_cost_width; ++d) out << 'o';
            for (; d<dollars; ++d) out << 'O';
            for (; d<BREITE+1; ++d) out << ' ';

            // print time:
            int zzzzs = int((BREITE*st->Value())/max_time + 0.5);
            out << st->Text() << ' ';
            for (d=0; d<zzzzs && d<mean_time_width; ++d) out << 'z';
            for (; d<zzzzs; ++d) out << 'Z';
            for (; d<BREITE+1; ++d) out << ' ';

            // print price:
            int pppps = int((BREITE*sp->Value())/max_price + 0.5);
            out << sp->Text() << ' ';
            for (d=0; d<pppps && d<mean_price_width; ++d) out << 'p';
            for (; d<pppps; ++d) out << 'P';
            out << '\n';

            if ((rest-1)   <= extraLastSum1 && rest!=1) { // nicht den letzten
                last_cost1 += sc->Value();
                last_time1 += st->Value();
            }
            if ((rest-1)<=extraLastSum2 && rest!=1) { // nicht den letzten
                last_cost2 += sc->Value();
                last_time2 += st->Value();
            }

            ++sc; ++st; ++sp; --rest;
        }

        out << '\n';

        size_t sums = cost_summen.size();
        assert(sums==time_summen.size());

        out         << "Gesamtpreis                       : " << preisString(overall_cost) << '\n';
        meanValue = 0;

        if (sums) {
            meanValue = overall_cost/sums;
            out     << "Durchschnitt (alle Werte)         : " << preisString(meanValue) << '\n';
            if (sums>2) {
                meanValue = (overall_cost-firstAndLast_cost)/(sums-2);
                out << "Durchschnitt (ohne ersten/letzten): " << preisString(meanValue) << '\n';
                meanValue = last_cost1/extraLastSum1;
                out << "Durchschnitt (der letzten " << rs::str::strf("%2i", extraLastSum1) << ")     : " << preisString(meanValue) << '\n';
                meanValue = last_cost2/extraLastSum2;
                out << "Durchschnitt (der letzten " << rs::str::strf("%2i", extraLastSum2) << ")     : " << preisString(meanValue) << '\n';
            }
        }

        out << '\n';
        out         << "Gesamtzeit                        : " << timeString(overall_time) << '\n';
        meanValue = 0;

        if (sums) {
            meanValue = overall_time/sums;
            out     << "Durchschnitt (alle Werte)         : " << timeString(meanValue) << '\n';
            if (sums>2) {
                meanValue = (overall_time-firstAndLast_time)/(sums-2);
                out << "Durchschnitt (ohne ersten/letzten): " << timeString(meanValue) << '\n';
                meanValue = last_time1/extraLastSum1;
                out << "Durchschnitt (der letzten " << rs::str::strf("%2i", extraLastSum1) << ")     : " << timeString(meanValue) << '\n';
                meanValue = last_time2/extraLastSum2;
                out << "Durchschnitt (der letzten " << rs::str::strf("%2i", extraLastSum2) << ")     : " << timeString(meanValue) << '\n';
            }
        }

        out << '\n';
        out         << "Kosten pro Minute                 : " << preisString((overall_cost*60)/overall_time) << '\n';
        meanValue = 0;
    }

#undef BREITE

    //  ----------------------------------------------------------------------------------------------------------------------------
    //      void printSums(ostream& out, const ConnectionList& connections, const string& summenName, sum_predicate predicate,
    //  ----------------------------------------------------------------------------------------------------------------------------
    void printSums(ostream& out, const ConnectionList& connections, const string& summenName, sum_predicate predicate,
                   prefix_writer writePrefix, double& meanValue, int extraLastSum1, int extraLastSum2, bool for_each_provider)
    {
        cout << "\n========================================================================================================================\n";
        if (for_each_provider) {
            set<string> providers;
            printSums_internal(out, connections, summenName, predicate, writePrefix, meanValue, extraLastSum1, extraLastSum2, 0, &providers);

            for (set<string>::const_iterator p=providers.begin(); p!=providers.end(); ++p) {
                string provider_summenName = summenName+" ("+*p+")";
                printSums_internal(out, connections, provider_summenName, predicate, writePrefix, meanValue, extraLastSum1, extraLastSum2, &*p, 0);
            }
        }
        else {
            printSums_internal(out, connections, summenName, predicate, writePrefix, meanValue, extraLastSum1, extraLastSum2, 0, 0);
        }
    }

    // -----------------------------------------------------------------------------------------------------------------------------------------
    //      void printCategorizedSums(ostream& out, const ConnectionList& connections, const string& summenName, sum_categorizer categorize,
    // -----------------------------------------------------------------------------------------------------------------------------------------
    void printCategorizedSums(ostream& out, const ConnectionList& connections, const string& summenName, sum_categorizer categorize,
                              category_prefix_writer writePrefix, bool for_each_provider)
    {
        cout << "\n========================================================================================================================\n";
        if (for_each_provider) {
            set<string> providers;
            printCategorizedSums_internal(out, connections, summenName, categorize, writePrefix, 0, &providers);

            for (set<string>::const_iterator p=providers.begin(); p!=providers.end(); ++p) {
                string provider_summenName = summenName+" ("+*p+")";
                printCategorizedSums_internal(out, connections, provider_summenName, categorize, writePrefix, &*p, 0);
            }
        }
        else {
            printCategorizedSums_internal(out, connections, summenName, categorize, writePrefix, 0, 0);
        }
    }


};
